﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;

using OldBvets.TrainDat;
using BvetsTool;

namespace BvetsTool
{
    using OldBvets.TrainDat;
    using Bvets5.Vehicle;
    using BvetsToolApi;


    public class PenColor
    {


    }

    public class DrawGraph
    {
        public enum ValueType
        {
            Acceleration,
            Force,
            Current,
        }

        Graphics g;

        private int[,] penColor = new int[8, 3]// グラフのペン色
        { { 255, 0, 0 }, { 255, 128, 0 }, { 255, 255, 0 }, { 0, 255, 0 },
        { 0, 255, 255 }, { 0, 0, 255 }, { 128, 0, 255}, { 255, 0, 255 }};

        private int width;
        private int height;

        // プロパティ
        public int MaxNotch { get; set; }
        public double XMax { get; set; }
        public double YMax { get; set; }
        public double XInterval { get; set; }
        public double YInterval { get; set; }
        public double YRatio { get; set; }
        public Color BackGroundColor { get; set; }
        public Color LineColor { get; set; }
        public Color FontColor { get; set; }

        // プロパティ化？
        // 色リスト
        // ペン太さ
        private Color[] colorList;
        private int penWeidht;
        private Point[] points;
        private Brush bgBrush;

        /// <summary>
        /// コンストラクタ
        /// </summary>
        /// <param name="bitmap"></param>
        public DrawGraph(Bitmap bitmap)
        {
            g = Graphics.FromImage(bitmap);
            width = bitmap.Width;
            height = bitmap.Height;

            this.MaxNotch = 8;
            this.XMax = 160;
            this.YMax = 4;
            this.XInterval = 10;
            this.YInterval = 1;
            this.YRatio = 0.8;
        }


        public void SizeChanged(Bitmap bitmap)
        {
            g = Graphics.FromImage(bitmap);
            width = bitmap.Width;
            height = bitmap.Height;
        }

        /// <summary>
        /// 描画領域左下を原点に変更
        /// </summary>
        /// <param name="g"></param>
        /// <param name="h"></param>
        private void graphicsTransform(Graphics g, int h)
        {
            g.ScaleTransform(1.0f, -1.0f);
            g.TranslateTransform(0, -h + 1);
            // gは参照渡しのためReturn不要
        }

        /// <summary>
        /// グラフのグリッド描画
        /// </summary>
        /// <param name="g"></param>
        /// <param name="bitmap"></param>
        /// <param name="pen"></param>
        private void drawGrid(Graphics g, Bitmap bitmap, Pen pen)
        {
            for (int i = 0; i <= XMax / XInterval; i++)
            {
                g.DrawLine(pen, i * (int)(width / XMax * XInterval), 0, i * (int)(width / XMax * XInterval), height - 1);
            }
            for (int i = 0; i <= YMax / YInterval; i++)
            {
                g.DrawLine(pen, 0, i * (int)(height / YMax * YInterval), width - 1, i * (int)(height / YMax * YInterval));
            }
        }

        /// <summary>
        /// グラフの目盛数値描画
        /// </summary>
        /// <param name="g"></param>
        /// <param name="bitmap"></param>
        /// <param name="font"></param>
        /// <param name="brush"></param>
        private void drawNumber(Graphics g, Bitmap bitmap, Font font, Brush brush)
        {
            // X
            for (int i = 0; i <= XMax * XInterval; i++)
            {
                g.ResetTransform();
                if (i * XInterval < 100) // 二桁
                {
                    g.DrawString(Convert.ToString(i * XInterval), font, brush, i * (int)(width / XMax * XInterval) - 7, height - 12);
                }
                else // 三桁
                {
                    g.DrawString(Convert.ToString(i * XInterval), font, brush, i * (int)(width / XMax * XInterval) - 10, bitmap.Height - 12);
                }
                graphicsTransform(g, bitmap.Height);
            }

            // Y
            for (int i = 0; i <= YMax / YInterval; i++)
            {
                g.ResetTransform();
                g.DrawString(Convert.ToString(i * YInterval), font, brush, -1, height - i * (int)(height / YMax * YInterval));
                graphicsTransform(g, height);
            }
        }

        /// <summary>
        /// グラフ下地描画
        /// </summary>
        /// <param name="bitmap"></param>
        public void DrawBase(Bitmap bitmap)
        {
            // 左下を原点に
            graphicsTransform(g, height);

            Pen pen = new Pen(Color.FromArgb(64, 64, 64)); // 線色(太さも？)
            Brush brush = new SolidBrush(Color.FromArgb(96, 96, 96)); // 文字色
            Font font = new Font("MS UI Gothic", 9); // 目盛数値フォント

            // 領域塗り潰し
            bgBrush = new SolidBrush(BackGroundColor);
            g.FillRectangle(bgBrush, g.VisibleClipBounds);

            // 原点変更
            graphicsTransform(g, bitmap.Height);
            
            // グリッド描画
            drawGrid(g, bitmap, pen);

            // 目盛数値描画
            drawNumber(g, bitmap, font, brush);
        }

        /// <summary>
        /// 曲線描画用の座標(P,Q)計算
        /// </summary>
        /// <param name="acp"></param>
        /// <param name="type"></param>
        /// <param name="subResistance"></param>
        /// <returns></returns>
        /// 
        /*
        public Point[][] CalcPoints(Acceleration[] acp, ValueType type, bool subResistance)
        {
            Point[][] point;

            if (subResistance) // 走行抵抗を引く
            {
                point = new Point[MaxNotch][];

                for (int i = 0; i < MaxNotch; i++)
                {
                    point[i] = new Point[width];

                    for (int j = 0; j < width; j++)
                    {
                        point[i][j].X = j;
                        switch (type)
                        {
                            case ValueType.Acceleration:
                                point[i][j].Y = toQ(OrigMath.CalcAccelerationCurve(acp[i], fromP(j)) - OrigMath.RunningResistanceAcceleration(fromP(j)));
                                break;
                            case ValueType.Force:
                                point[i][j].Y = toQ(OrigMath.ForceFromAcceleration(OrigMath.CalcAccelerationCurve(acp[i], fromP(j))) - (OrigMath.RunningResistanceForce(fromP(j)) / Parameters.Dynamics.MotorcarCount));
                                break;
                            case ValueType.Current:
                                point[i][j].Y = toQ(OrigMath.CalcCurrentCurve(acp[i], fromP(j)) / maxAcceleration(acp) * 100);
                                break;
                        }
                    }
                }
            }
            else // 走行抵抗を引かない
            {
                point = new Point[MaxNotch + 1][];

                for (int i = 0; i < MaxNotch + 1; i++) // 各曲線
                {
                    point[i] = new Point[width];

                    for (int j = 0; j < width; j++) // 各PからQを計算
                    {
                        point[i][j].X = j;

                        if (i == 0) // 走行抵抗曲線
                        {
                            switch (type)
                            {
                                case ValueType.Acceleration:
                                    point[i][j].Y = toQ(OrigMath.RunningResistanceAcceleration(fromP(j)));
                                    break;
                                case ValueType.Force:
                                    point[i][j].Y = toQ(OrigMath.RunningResistanceForce(fromP(j)) / Parameters.Dynamics.MotorcarCount);
                                    break;
                                case ValueType.Current:
                                    point[i][j].Y = 0;
                                    break;
                            }
                        }
                        else // ノッチ曲線
                        {
                            switch (type)
                            {
                                case ValueType.Acceleration:
                                    point[i][j].Y = toQ(OrigMath.CalcAccelerationCurve(acp[i - 1], fromP(j)));
                                    break;
                                case ValueType.Force:
                                    point[i][j].Y = toQ(OrigMath.ForceFromAcceleration(OrigMath.CalcAccelerationCurve(acp[i - 1], fromP(j))));
                                    break;
                                case ValueType.Current:
                                    point[i][j].Y = toQ(OrigMath.CalcCurrentCurve(acp[i - 1], fromP(j)) / maxAcceleration(acp) * 100);
                                    break;
                            }
                        }
                    }

                }
            }

            return point;
        }
        */

        /// <summary>
        /// 加速度パラメーターから座標テーブルを作成
        /// </summary>
        /// <param name="acp"></param>
        /// <param name="type"></param>
        /// <param name="subResistance"></param>
        /// <returns></returns>
        public Point[][] CalcPoints(Acceleration[] acp, ValueType type, bool subResistance)
        {
            Point[][] point;

            point = new Point[MaxNotch + 1][];

            for (int i = 0; i < MaxNotch + 1; i++)
            {
                point[i] = new Point[width];

                for (int j = 0; j < width; j++)
                {
                    // X座標を設定
                    point[i][j].X = j;

                    // Y座標を設定
                    if (subResistance) // 走行抵抗を引く
                    {
                        if (i == 0) // 空の走行抵抗曲線
                        {
                            point[i][j].Y = 0;
                        }
                        else // ノッチ曲線
                        {
                            switch (type)
                            {
                                case ValueType.Acceleration:
                                    point[i][j].Y = toQ(OrigMath.CalcAccelerationCurve(acp[i - 1], fromP(j)) - OrigMath.RunningResistanceAcceleration(fromP(j)));
                                    break;
                                case ValueType.Force:
                                    point[i][j].Y = toQ(OrigMath.ForceFromAcceleration(OrigMath.CalcAccelerationCurve(acp[i - 1], fromP(j))) - (OrigMath.RunningResistanceForce(fromP(j)) / Parameters.Dynamics.MotorcarCount));
                                    break;
                                case ValueType.Current:
                                    point[i][j].Y = toQ(OrigMath.CalcCurrentCurve(acp[i - 1], fromP(j)) / MaxAcceleration(acp) * 100);
                                    break;
                            }
                        }
                    }
                    else // 走行抵抗を表示
                    {
                        if (i == 0) // 走行抵抗曲線
                        {
                            switch (type)
                            {
                                case ValueType.Acceleration:
                                    point[i][j].Y = toQ(OrigMath.RunningResistanceAcceleration(fromP(j)));
                                    break;
                                case ValueType.Force:
                                    point[i][j].Y = toQ(OrigMath.RunningResistanceForce(fromP(j)) / Parameters.Dynamics.MotorcarCount);
                                    break;
                                case ValueType.Current:
                                    point[i][j].Y = 0;
                                    break;
                            }
                        }
                        else // ノッチ曲線
                        {
                            switch (type)
                            {
                                case ValueType.Acceleration:
                                    point[i][j].Y = toQ(OrigMath.CalcAccelerationCurve(acp[i - 1], fromP(j)));
                                    break;
                                case ValueType.Force:
                                    point[i][j].Y = toQ(OrigMath.ForceFromAcceleration(OrigMath.CalcAccelerationCurve(acp[i - 1], fromP(j))));
                                    break;
                                case ValueType.Current:
                                    point[i][j].Y = toQ(OrigMath.CalcCurrentCurve(acp[i - 1], fromP(j)) / MaxAcceleration(acp) * 100);
                                    break;
                            }
                        }
                    }
                }
            }

            return point;
        }

        /// <summary>
        /// テーブルからグラフ描画用座標を作成
        /// </summary>
        /// <param name="table"></param>
        /// <param name="type"></param>
        /// <param name="subResistance"></param>
        /// <returns></returns>
        public Point[][] CalcPoints(double[][,] table, ValueType type, bool subResistance)
        {
            Point[][] point;


            point = new Point[table.Length + 1][];

            for (int i = 0; i < table.Length + 1; i++)
            {
                int length;

                if (i == 0)
                {
                    length = table[i].GetLength(0);
                }
                else
                {
                    length = table[i - 1].GetLength(0);
                }

                point[i] = new Point[(int)(length)];

                for (int j = 0; j < length; j++)
                {
                    double x;

                    // X座標を設定
                    if (i == 0)
                    {
                        x = table[i][j, 0];
                    }
                    else
                    {
                        x = table[i - 1][j, 0];
                    }

                    point[i][j].X = (int)(x * ((double)width / XMax));

                    // Y座標を設定
                    if (subResistance) // 走行抵抗を引く
                    {
                        if (i == 0) // 空の走行抵抗曲線
                        {
                            point[i][j].Y = 0;
                        }
                        else // ノッチ曲線
                        {
                            switch (type)
                            {
                                case ValueType.Acceleration:
                                    point[i][j].Y = toQ(OrigMath.MinAccelerationFromForce(table[i - 1][j, 1]) - OrigMath.RunningResistanceAcceleration(x));
                                    break;
                                case ValueType.Force:
                                    point[i][j].Y = toQ(table[i - 1][j, 1] - OrigMath.RunningResistanceForce(x) / Parameters.Dynamics.MotorcarCount);
                                    break;
                                case ValueType.Current:
                                    point[i][j].Y = toQ(table[i - 1][j, 1]);
                                    break;
                            }
                        }
                    }
                    else // 走行抵抗を表示
                    {
                        if (i == 0) // 走行抵抗曲線
                        {
                            switch (type)
                            {
                                case ValueType.Acceleration:
                                    point[i][j].Y = toQ(OrigMath.RunningResistanceAcceleration(x));
                                    break;
                                case ValueType.Force:
                                    point[i][j].Y = toQ(OrigMath.RunningResistanceForce(x) / Parameters.Dynamics.MotorcarCount);
                                    break;
                                case ValueType.Current:
                                    point[i][j].Y = 0;
                                    break;
                            }
                        }
                        else // ノッチ曲線
                        {
                            switch (type)
                            {
                                case ValueType.Acceleration:
                                    point[i][j].Y = toQ(OrigMath.MinAccelerationFromForce(table[i - 1][j, 1]));
                                    break;
                                case ValueType.Force:
                                    point[i][j].Y = toQ(table[i - 1][j, 1]);
                                    break;
                                case ValueType.Current:
                                    point[i][j].Y = toQ(table[i - 1][j, 1]);
                                    break;
                            }
                        }
                    }
                }
            }

            return point;
        }
        /// <summary>
        /// 最大加速度（相対電流計算用）
        /// </summary>
        /// <param name="acp"></param>
        /// <returns></returns>
        public double MaxAcceleration(Acceleration[] acp)
        {
            double max = acp[0].MaxAcceleration;

            for (int i = 0; i < acp.Length - 1; i++)
            {
                if (acp[i].MaxAcceleration < acp[i + 1].MaxAcceleration)
                {
                    max = acp[i + 1].MaxAcceleration;
                }
            }

            return max;
        }

        /// <summary>
        /// 曲線描画用のペン色設定
        /// </summary>
        /// <param name="index"></param>
        /// <param name="subResistance"></param>
        /// <returns></returns>
        public Pen[] SetPenColor(int index, bool subResistance, double rate)
        {
            Pen[] pen = new Pen[MaxNotch + 1];

            pen = new Pen[MaxNotch + 1];

            if (subResistance) // 走行抵抗を引く
            {
                pen[0] = new Pen(Color.FromArgb(0, 0, 0, 0));
            }
            else // 走行抵抗を表示
            {
                pen[0] = new Pen(Color.FromArgb(128, 128, 128));
            }

            for (int i = 0; i < MaxNotch; i++)
            {
                int col = i % 8;

                if (i == index) // 選択中ノッチ曲線
                {
                    pen[i + 1] = new Pen(Color.FromArgb(penColor[col, 0], penColor[col, 1], penColor[col, 2]));
                }
                else // それ以外のノッチ曲線
                {
                    pen[i + 1] = new Pen(OrigMath.NonseledtedIndexColor(Color.FromArgb(penColor[col, 0], penColor[col, 1], penColor[col, 2]), rate));
                }
            }

            return pen;
        }

        /// <summary>
        /// 座標上の数値Pから実際の数値Xへ
        /// </summary>
        /// <param name="p">P</param>
        /// <returns>X</returns>
        private double fromP(int p)
        {
            return (double)p / (double)width * XMax;
        }

        /// <summary>
        /// 実際の数値Yから座標上の数値Qへ
        /// </summary>
        /// <param name="y">Y</param>
        /// <returns>Q</returns>
        private int toQ(double y)
        {
            return (int)(y / (double)YMax * height);
        }

        /// <summary>
        /// 曲線描画
        /// </summary>
        /// <param name="pen"></param>
        /// <param name="point"></param>
        /// <param name="index"></param>
        /// <param name="subResistance"></param>
        public void DrawCurves(Pen[] pen, Point[][] point,int index)
        {
            // 例外判定
            if (pen == null | point == null) // ペン・座標オブジェクトが空の場合
            {
                return;
            }
            else if (point[0].Length < 2) // 座標が一つ以下の場合（直線が引けない）
            {
                return;
            }
            else if (point.Length != pen.Length) // ペンの配列数と座標組の配列数が一致しない場合 
            {
                return;
            }

            // ペン（座標）の数だけ曲線描画
            for (int i = 0; i < point.Length; i++)
            {
                g.DrawLines(pen[i], point[i]);
            }
            g.DrawLines(pen[index + 1], point[index + 1]);
        }

        public void DrawCurves(Pen[] pen, PointF[][] point, int index)
        {
            for (int i = 0; i < point.Length; i++)
            {
                g.DrawLines(pen[i], point[i]);
            }
            g.DrawLines(pen[index + 1], point[index + 1]);
        }

        /// <summary>
        /// マウスカーソルX座標を実数値に変換
        /// </summary>
        /// <param name="x">カーソルX座標</param>
        /// <returns></returns>
        public double PointX(double x)
        {
            return x / width * XMax;
        }

        /// <summary>
        /// マウスカーソルX座標を実数値に変換
        /// </summary>
        /// <param name="y">カーソルY座標</param>
        /// <returns></returns>
        public double PointY(double y)
        {
            return (height - 1 - y) / (double)height * YMax;
        }






        // public void DrawCurve(Bitmap bitmap, Pen pen, Point[] point)
        // 曲線描画

        /*
        // pen切り替え無し（遅い）
        for (int x = 0; x < bitmap.Width; x++)
        {
            // 走行抵抗 // 'x', pen, Width, Height, XMax, YMax
            pen.Color = Color.FromArgb(128, 128, 128);
            double v = x / (double)bitmap.Width * XMax; // (double)無しの場合0に
            int y0 = (int)(bitmap.Height - 1 - bitmap.Height / YMax * BvetsToolApi.OrigMath.AccelerationFromForce(BvetsToolApi.OrigMath.RunningResistanceForce(v)));
            int y1 = (int)(bitmap.Height - 1 - bitmap.Height / YMax * BvetsToolApi.OrigMath.AccelerationFromForce(BvetsToolApi.OrigMath.RunningResistanceForce(v + 1)));
            g.DrawLine(pen, x, y0, x + 1, y1);

            // その他のノッチ
            for (int i = 0; i < MaxNotch; i++)
            {
                pen.Color = Color.FromArgb(penColor[i, 0] / 2, penColor[i, 1] / 2, penColor[i, 2] / 2);
                int y4 = (int)CalcNotchCurve(accParam[i], x, bitmap.Width, bitmap.Height);
                int y5 = (int)CalcNotchCurve(accParam[i], x + 1, bitmap.Width, bitmap.Height);
            g.DrawLine(pen, x, y4, x + 1, y5);
            }
                
            // 選択中ノッチ
            pen.Color = Color.FromArgb(penColor[index, 0], penColor[index, 1], penColor[index, 2]);
            int y2 = (int)CalcNotchCurve(accParam[index], x, bitmap.Width, bitmap.Height);
            int y3 = (int)CalcNotchCurve(accParam[index], x + 1, bitmap.Width, bitmap.Height);
            g.DrawLine(pen, x, y2, x + 1, y3);
        }
        */

        // pen切り替え無し（高速）
        /*
        // その他のノッチ
        for (int i = 0; i < MaxNotch; i++)
        {
            pen.Color = Color.FromArgb(penColor[i, 0] / 2, penColor[i, 1] / 2, penColor[i, 2] / 2);
            for (int x = 0; x < bitmap.Width; x++)
            {
                int y0 = (int)CalcNotchCurve(accParam[i], x, bitmap.Width, bitmap.Height);
                int y1 = (int)CalcNotchCurve(accParam[i], x + 1, bitmap.Width, bitmap.Height);
                g.DrawLine(pen, x, y0, x + 1, y1);
                // g.FillRectangle(Brushes.Red, new Rectangle(x, y0, 1, Math.Abs(y1 - y0 + 1)));
            }
        }

        // 選択中ノッチ
        for (int x = 0; x < bitmap.Width; x++)
        {
            pen.Color = Color.FromArgb(penColor[index, 0], penColor[index, 1], penColor[index, 2]);
            
            int y0 = (int)CalcNotchCurve(accParam[index], x, bitmap.Width, bitmap.Height);
            int y1 = (int)CalcNotchCurve(accParam[index], x + 1, bitmap.Width, bitmap.Height);
            g.DrawLine(pen, x, y0, x + 1, y1);
            // g.FillRectangle(Brushes.Red, new Rectangle(x, y0, 1, Math.Abs(y1 - y0 + 1)));
            
        // 走行抵抗
            pen.Color = Color.FromArgb(128, 128, 128);

            double v = x / (double)bitmap.Width * XMax; // (double)無しの場合0に
            int y2 = (int)(bitmap.Height - 1 - bitmap.Height / YMax * BvetsToolApi.OrigMath.AccelerationFromForce(BvetsToolApi.OrigMath.RunningResistanceForce(v)));
            int y3 = (int)(bitmap.Height - 1 - bitmap.Height / YMax * BvetsToolApi.OrigMath.AccelerationFromForce(BvetsToolApi.OrigMath.RunningResistanceForce(v + 1)));
            g.DrawLine(pen, x, y2, x + 1, y3);
            // g.FillRectangle(Brushes.Red, new Rectangle(x, y0, 1, Math.Abs(y1 - y0 + 1)));
        }
        */

        // 終了処理
        // brush.Dispose();
        // pen.Dispose();
        // g.Dispose();
        //}
        
        public double CalcNotchCurve(Acceleration acp, double x, int width, int height)
        {
            double v = x / width * XMax;
            double y;

            if (v < acp.ConstPowerStart) // 定引張力域
            {
                y = BvetsToolApi.OrigMath.LinearInterpolation(v, 0.001, acp.ConstAccelerationStart, acp.ConstPowerStart, acp.ConstAccelerationEnd);
            }
            else if (v < acp.ConstPowerEnd) // 定出力域, 0.001はゼロ除算防止
            {
                y = acp.ConstAccelerationEnd * acp.ConstPowerStart / (v + 0.001);
            }
            else // 特性域, 0.001はゼロ除算防止
            {
                y = acp.ConstAccelerationEnd * acp.ConstPowerStart / (acp.ConstPowerEnd + 0.001) * Math.Pow((acp.ConstPowerEnd + 0.001) / (v + 0.001), acp.CharacteristicCurveDegree);
            }
            return height - (1 + y * height / YMax);
        }
    }
}