﻿// INI形式のファイルを読む
// http://smdn.jp/programming/netfx/tips/read_ini/

using System;
using System.Collections.Generic;
using System.IO;
using System.Text.RegularExpressions;

using Bvets5.Vehicle;

class IniReader
{
    public static void IniParser(string filepath)
    {
        string iniFile = filepath;

        // INIファイルを読み込み、各セクション毎に表示する
        foreach (var section in ReadIni(iniFile))
        {
            // Console.WriteLine("[{0}]", section.Key);
            // Dynamicsセクションから読み込む
            if (section.Key.ToLower() == "dynamics")
            {
                foreach (var pair in section.Value)
                {
                    double value;
                    switch (pair.Key.ToLower())
                    {
                        case "motorcarweight":
                            if (double.TryParse(pair.Value, out value))
                            {
                                Parameters.Dynamics.MotorcarWeight = value;
                            }
                            else
                            {
                                Parameters.Dynamics.MotorcarWeight = 31500;
                            }
                            break;
                        case "motorcarcount":
                            if (double.TryParse(pair.Value, out value))
                            {
                                Parameters.Dynamics.MotorcarCount = value;
                            }
                            else
                            {
                                Parameters.Dynamics.MotorcarCount = 5;
                            }
                            break;
                        case "trailerweight":
                            if (double.TryParse(pair.Value, out value))
                            {
                                Parameters.Dynamics.TrailerWeight = value;
                            }
                            else
                            {
                                Parameters.Dynamics.TrailerWeight = 31500;
                            }
                            break;
                        case "trailercount":
                            if (double.TryParse(pair.Value, out value))
                            {
                                Parameters.Dynamics.TrailerCount = value;
                            }
                            else
                            {
                                Parameters.Dynamics.TrailerCount = 5;
                            }
                            break;
                        case "motorcarinertiafactor":
                            if (double.TryParse(pair.Value, out value))
                            {
                                Parameters.Dynamics.MotorcarInertiaFactor = value;
                            }
                            else
                            {
                                Parameters.Dynamics.MotorcarInertiaFactor = 0.1;
                            }
                            break;
                        case "trailerinertiafactor":
                            if (double.TryParse(pair.Value, out value))
                            {
                                Parameters.Dynamics.TrailerInertiaFactor = value;
                            }
                            else
                            {
                                Parameters.Dynamics.TrailerInertiaFactor = 0.05;
                            }
                            break;
                    }
                    // Console.WriteLine("  <{0}>='{1}'", pair.Key, pair.Value);
                }
            }
        }
    }

    private static Dictionary<string, Dictionary<string, string>> ReadIni(string file)
    {
        using (var reader = new StreamReader(file))
        {
            var sections = new Dictionary<string, Dictionary<string, string>>(StringComparer.Ordinal);
            var regexSection = new Regex(@"^\s*\[(?<section>[^\]]+)\].*$", RegexOptions.Singleline | RegexOptions.CultureInvariant);
            var regexNameValue = new Regex(@"^\s*(?<name>[^=]+)=(?<value>.*?)(\s+;(?<comment>.*))?$", RegexOptions.Singleline | RegexOptions.CultureInvariant);
            var currentSection = string.Empty;

            // セクション名が明示されていない先頭部分のセクション名を""として扱う
            sections[string.Empty] = new Dictionary<string, string>();

            for (; ; )
            {
                var line = reader.ReadLine();

                if (line == null)
                    break;

                // 空行は読み飛ばす
                if (line.Length == 0)
                    continue;

                // コメント行は読み飛ばす
                if (line.StartsWith(";", StringComparison.Ordinal))
                    continue;
                else if (line.StartsWith("#", StringComparison.Ordinal))
                    continue;

                var matchNameValue = regexNameValue.Match(line);

                if (matchNameValue.Success)
                {
                    // name=valueの行
                    sections[currentSection][matchNameValue.Groups["name"].Value.Trim()] = matchNameValue.Groups["value"].Value.Trim();
                    continue;
                }

                var matchSection = regexSection.Match(line);

                if (matchSection.Success)
                {
                    // [section]の行
                    currentSection = matchSection.Groups["section"].Value;

                    if (!sections.ContainsKey(currentSection))
                        sections[currentSection] = new Dictionary<string, string>();

                    continue;
                }
            }

            return sections;
        }
    }
}
